## Turtle graphics API for Ripen Forth using NiGui.

import math
import nigui
import ripen

var image* = newImage() ## Internal buffer for drawing operations.

proc setupImage*() =
        image.resize 640, 400
        image.canvas.areaColor = rgb(0, 0, 0)
        image.canvas.lineColor = rgb(255, 255, 255)
        image.canvas.textColor = rgb(255, 255, 255)
        image.canvas.fill

setupImage()

var turtleX* = float(image.width div 2) ## Starting in the upper-left corner.
var turtleY* = float(image.height div 2) ## Y grows down as usual today.
var heading* = 0.0 ## In degrees, with zero being up.

var showTurtle* = true
var isPenDown* = true

var modified* = false ## For use in the GUI; never gets reset here.

proc drawLineRelative*(canvas: Canvas; x, y: int; angle, length: float) =
        let angle = degToRad(angle - 90.0)
        let x2 = x + int(cos(angle) * length)
        let y2 = y + int(sin(angle) * length)
        canvas.drawLine x, y, x2, y2

proc centerImage*(dest: Canvas, source: Image): (int, int) =
        let xOff = (dest.width - source.width) div 2
        let yOff = (dest.height - source.height) div 2

        dest.areaColor = source.canvas.areaColor
        dest.fill
        dest.drawImage(source, xOff, yOff)

        return (xOff, yOff)

var palette* = [
        rgb(  0,   0,   0),
        rgb(  0,   0, 255),
        rgb(  0,  80,   0),
        rgb(  0, 255, 255),
        rgb(255,   0,   0),
        rgb(255,   0, 255),
        rgb(255, 255,   0),
        rgb(255, 255, 255),
        rgb(165,  42,  42),
        rgb(210, 180, 140),
        rgb( 34, 139,  34),
        rgb(127, 255, 212), # Using the value for aquamarine, not aqua.
        rgb(250, 128, 114),
        rgb(128,   0, 128),
        rgb(255, 165,   0),
        rgb(128, 128, 128)
]

var colors* = {
        "black": rgb(  0,   0,   0),
        "blue": rgb(  0,   0, 255),
        "green": rgb(  0,  80,   0),
        "cyan": rgb(  0, 255, 255),
        "red": rgb(255,   0,   0),
        "magenta": rgb(255,   0, 255),
        "yellow": rgb(255, 255,   0),
        "white": rgb(255, 255, 255),
        "brown": rgb(165,  42,  42),
        "tan": rgb(210, 180, 140),
        "forest": rgb( 34, 139,  34), # Using the value for forest green.
        "aqua": rgb(127, 255, 212), # Using the value for aquamarine.
        "salmon": rgb(250, 128, 114),
        "purple": rgb(128,   0, 128),
        "orange": rgb(255, 165,   0),
        "grey": rgb(128, 128, 128)
}.toTable

proc forward() =
        let angle = degToRad(heading - 90.0)
        let distance = stack.pop
        let tx = int(turtleX)
        let ty = int(turtleY)
        let x = turtleX + cos(angle) * distance
        let y = turtleY + sin(angle) * distance
        if isPenDown:
                image.canvas.drawLine tx, ty, int(x), int(y)
                modified = true
        turtleX = x
        turtleY = y

proc back() =
        let angle = degToRad(heading - 90.0)
        let distance = stack.pop
        let tx = int(turtleX)
        let ty = int(turtleY)
        let x = turtleX - cos(angle) * distance
        let y = turtleY - sin(angle) * distance
        if isPenDown:
                image.canvas.drawLine tx, ty, int(x), int(y)
                modified = true
        turtleX = x
        turtleY = y

proc left() =
        heading -= stack.pop
        if heading < 0: heading += 360

proc right() =
        heading += stack.pop
        if heading < 0: heading -= 360

proc setheading() = heading = stack.pop mod 360

words["forward"] = primitive(forward)
words["back"] = primitive(back)
words["left"] = primitive(left)
words["right"] = primitive(right)
words["setheading"] = primitive(setheading)

proc setxy() =
        let tx = int(turtleX)
        let ty = int(turtleY)
        let y = stack.pop
        let x = stack.pop
        if isPenDown:
                image.canvas.drawLine tx, ty, int(x), int(y)
                modified = true
        turtleX = x
        turtleY = y

proc setx() =
        let tx = int(turtleX)
        let ty = int(turtleY)
        let x = stack.pop
        if isPenDown:
                image.canvas.drawLine tx, ty, int(x), ty
                modified = true
        turtleX = x

proc sety() =
        let tx = int(turtleX)
        let ty = int(turtleY)
        let y = stack.pop
        if isPenDown:
                image.canvas.drawLine tx, ty, tx, int(y)
                modified = true
        turtleY = y

words["setxy"] = primitive(setxy)
words["setx"] = primitive(setx)
words["sety"] = primitive(sety)

proc home() =
        let y = image.height div 2
        let x = image.width div 2
        if isPenDown:
                image.canvas.drawLine int(turtleX), int(turtleY), x, y
                modified = true
        turtleX = float(x)
        turtleY = float(y)
        heading = 0.0

words["home"] = primitive(home)

proc pos() =
        stack.add turtleX
        stack.add turtleY
proc xcor() = stack.add turtleX
proc ycor() = stack.add turtleY
proc pheading() = stack.add heading
proc towards() =
        let dy = stack.pop() - turtleY
        let dx = stack.pop() - turtleX
        stack.add arctan2(dy, dx).radToDeg + 90

words["pos"] = primitive(pos)
words["xcor"] = primitive(xcor)
words["ycor"] = primitive(ycor)
words["heading"] = primitive(pheading)
words["towards"] = primitive(towards)

proc penup() = isPenDown = false
proc pendown() = isPenDown = true
proc pendownp() = stack.add float(isPenDown)

words["penup"] = primitive(penup)
words["pendown"] = primitive(pendown)
words["pendown?"] = primitive(pendownp)

proc hideturtle() = showTurtle = false
proc pshowturtle() = showTurtle = true
proc shownp() = stack.add float(showTurtle)

words["hideturtle"] = primitive(hideturtle)
words["showturtle"] = primitive(pshowturtle)
words["shown?"] = primitive(shownp)

proc setpensize() = image.canvas.lineWidth = stack.pop
proc pensize() = stack.add image.canvas.lineWidth

words["setpensize"] = primitive(setpensize)
words["pensize"] = primitive(pensize)

proc clean() = image.canvas.fill
proc clearscreen() =
        image.canvas.fill
        modified = true
        turtleX = float(image.width div 2)
        turtleY = float(image.height div 2)
        heading = 0.0

words["clean"] = primitive(clean)
words["clearscreen"] = primitive(clearscreen)

proc ellipse() =
        let ry = int(stack.pop)
        let rx = int(stack.pop)
        let y = int(turtleY) - ry
        let x = int(turtleX) - rx
        modified = true
        image.canvas.drawEllipseOutline x, y, rx * 2, ry * 2
proc fillellipse() =
        let ry = int(stack.pop)
        let rx = int(stack.pop)
        let y = int(turtleY) - ry
        let x = int(turtleX) - rx
        modified = true
        image.canvas.drawEllipseArea x, y, rx * 2, ry * 2

words["ellipse"] = primitive(ellipse)
words["fillellipse"] = primitive(fillellipse)

proc label() =
        modified = true
        image.canvas.drawText scratch, int(turtleX), int(turtleY)
proc setlabelheight() = image.canvas.fontSize = stack.pop
proc labelsize() =
        stack.add image.canvas.getTextWidth(scratch).toFloat
        stack.add image.canvas.getTextLineHeight.toFloat

words["label"] = primitive(label)
words["setlabelheight"] = primitive(setlabelheight)
words["labelsize"] = primitive(labelsize)

proc setpencolor() =
        let blue = byte(stack.pop)
        let green = byte(stack.pop)
        let red = byte(stack.pop)
        image.canvas.lineColor = rgb(red, green, blue)
        image.canvas.textColor = image.canvas.lineColor
proc pencolor() =
        var pc = image.canvas.lineColor
        stack.add float(pc.red)
        stack.add float(pc.green)
        stack.add float(pc.blue)
proc setbackground() =
        let blue = byte(stack.pop)
        let green = byte(stack.pop)
        let red = byte(stack.pop)
        image.canvas.areaColor = rgb(red, green, blue)
proc background() =
        var bg = image.canvas.areaColor
        stack.add float(bg.red)
        stack.add float(bg.green)
        stack.add float(bg.blue)
proc ppalette() =
        var pc = palette[int(stack.pop)]
        stack.add float(pc.red)
        stack.add float(pc.green)
        stack.add float(pc.blue)
proc setpalette() =
        let entry = int(stack.pop)
        let blue = byte(stack.pop)
        let green = byte(stack.pop)
        let red = byte(stack.pop)
        palette[entry] = rgb(red, green, blue)

words["setpencolor"] = primitive(setpencolor)
words["pencolor"] = primitive(pencolor)
words["setbackground"] = primitive(setbackground)
words["background"] = primitive(background)
words["palette"] = primitive(ppalette)
words["setpalette"] = primitive(setpalette)

sigil['#'] = proc(n: string) =
        var pc = colors[n]
        stack.add float(pc.red)
        stack.add float(pc.green)
        stack.add float(pc.blue)

alias["fd"] = "forward"
alias["bk"] = "back"
alias["lt"] = "left"
alias["rt"] = "right"
alias["seth"] = "setheading"

alias["ht"] = "hideturtle"
alias["st"] = "showturtle"
alias["pu"] = "penup"
alias["pd"] = "pendown"
alias["cs"] = "clearscreen"

alias["setpc"] = "setpencolor"
alias["pc"] = "pencolor"
alias["setbg"] = "setbackground"
alias["bg"] = "background"
